/* $USAGI: sa_index.c,v 1.13 2001/10/10 07:37:05 miyazawa Exp $ */

/*
 * Copyright (C) 2001 USAGI/WIDE Project.
 * All rights reserved.
 * 
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 3. Neither the name of the project nor the names of its contributors
 *    may be used to endorse or promote products derived from this software
 *    without specific prior written permission.
 * 
 * THIS SOFTWARE IS PROVIDED BY THE PROJECT AND CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL THE PROJECT OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 * 
 */

/*
 * struct ipsec_sa is connected with struct ipsec_sp by struct sa_index
 * The element sa of sa_index occasionlly NULL.
 * Note that struct ipsec_sa has a reference count. 
 */


#ifdef MODULE
#  include <linux/module.h>
#  ifdef MODVERSIONS
#    include <linux/modversions.h>
#  endif /* MODVERSIONS */
#endif /* MODULE */

#ifdef SPD_DEBUG
#define DEBUG 1
#endif

#include <linux/config.h>
#include <linux/kernel.h>
#include <linux/version.h>
#include <linux/init.h>
#include <linux/net.h>
#include <linux/in.h>
#include <linux/in6.h>
#include <linux/malloc.h>
#include <linux/spinlock.h>
#include <linux/list.h>

#include <net/sadb.h>

#include "sadb_utils.h"

struct sa_index* sa_index_kmalloc()
{
	struct sa_index* sa_idx = NULL;

	sa_idx = kmalloc(sizeof(struct sa_index), GFP_KERNEL);

	if (!sa_idx) {
		pr_debug("sa_list_kmalloc: kmalloc faild\n");
		return NULL;
	}

	sa_index_init(sa_idx);

	return sa_idx;
}

int sa_index_init(struct sa_index *sa)
{
	int error = 0;

	if (!sa) {
		pr_debug("sa_index_init: sa is null\n");
		error = -EINVAL;
		goto err;
	}

	memset(sa, 0, sizeof(struct sa_index));
	INIT_LIST_HEAD(&(sa->entry));

err:
	return error;
}

int sa_index_kfree(struct sa_index **sa_idx)
{
	int error = 0;
	if (!*sa_idx) {
		pr_debug("sa_list_kfree: sa is null\n");
		error = -EINVAL;
		goto err;
	}

	if ((*sa_idx)->sa) {
		sadb_release_ref((*sa_idx)->sa);
		pr_debug("sa_index_kfree: ptr=%p,refcnt=%d\n",
			(*sa_idx)->sa, atomic_read(&(*sa_idx)->sa->refcnt));
		(*sa_idx)->sa = NULL;
	}

	kfree(*sa_idx);
err:
	return error;
}

int sa_index_copy(struct sa_index *dst, struct sa_index *src)
{
	int error = 0;

	if (!(dst&&src)) {
		pr_debug("sa_list_copy: dst or src is null\n");
		error = -EINVAL;
		goto err;
	}

	dst->dst	 = src->dst;
	dst->ipsec_proto = src->ipsec_proto;
	dst->spi 	 = src->spi;
	if (src->sa) {
		dst->sa		 = src->sa;
		atomic_inc(&(dst->sa->refcnt));
		pr_debug("sa_index_kfree: ptr=%p,refcnt=%d\n",
			dst->sa, atomic_read(&(dst->sa->refcnt)));
	}

err:
	return error;
}

int sa_index_copy_from_sa(struct sa_index *dst, struct ipsec_sa *src)
{
	int error = 0;

	if (!(dst&&src)) {
		pr_debug("sa_index_copy_from_sa: dst or src is null\n");
		error = -EINVAL;
		goto err;
	}

	dst->spi = src->spi;
	dst->ipsec_proto = src->ipsec_proto;

	switch (src->dst.ss_family) {

	case AF_INET:
		memcpy(&dst->dst, &src->dst, sizeof(struct sockaddr_in));
		break;
	
	case AF_INET6:
		memcpy(&dst->dst, &src->dst, sizeof(struct sockaddr_in6));
		break;

	default:
		error = -EINVAL;
		goto err;
	}

err:
	return error;
}

int sa_index_compare(struct sa_index *sa_idx1, struct sa_index *sa_idx2)
{
	if (!(sa_idx1&&sa_idx2)) {
		pr_debug("sa_index_compare_sa: sa_idx1 or sa_idx2 is null\n");
		return -EINVAL;
	}

	if (sa_idx1->spi != sa_idx2->spi) {
		return 1;
	}

	if (sa_idx1->ipsec_proto != sa_idx2->ipsec_proto) {
		return 1;
	}

	return compare_address_with_prefix((struct sockaddr*)&(sa_idx1->dst), sa_idx1->prefixlen_d,
				(struct sockaddr*)&(sa_idx2->dst), sa_idx2->prefixlen_d); 

}

int sa_list_append(struct list_head *list, struct sa_index *sa)
{
	int error = 0;
	struct sa_index *new = NULL;

	if (!(list&&sa)) {
		pr_debug("sa_list_append: list or sa is null\n");
		error = -EINVAL;
		goto err;
	}

	new = sa_index_kmalloc();
	if (!new) {
		pr_debug("sa_index_append: sa_index_kmalloc faild\n");
		error = -ENOMEM;
		goto err;
	}
	error = sa_index_init(new);
	if (error) {
		pr_debug("sa_index_append: sa_index_init failed\n");
		goto err;
	}
	error = sa_index_copy(new, sa);
	if (error) {
		pr_debug("sa_list_append: sa_index_copy failed\n");
		goto err;
	}
	list_add_tail(&(new->entry), list);

	pr_debug("sa_list_append: list->prev=%p\n", list->prev);
	pr_debug("sa_list_append: list->next=%p\n", list->next);

err:
	return error;
}

int sa_list_copy(struct list_head *dst, struct list_head *src)
{
	int error = 0;
	struct list_head *pos = NULL;
	struct sa_index *tmp = NULL, *new = NULL;

	if (!(dst&&src)) {
		pr_debug("sa_list_copy: dst or src is null\n");
		error = -EINVAL;
		goto err;
	}


	list_for_each(pos, src) {
		tmp = list_entry(pos, struct sa_index, entry);
		new = NULL;
		new = sa_index_kmalloc();
		if (!new) {
			pr_debug("sa_list_copy: sa_index_kmalloc failed\n");
			error = -ENOMEM;
			goto err;
		}
		error = sa_index_init(new);
		if (error) {
			pr_debug("sa_list_copy: sa_index_init failed\n");
			goto err;
		}
		error = sa_index_copy(new, tmp);
		if (error) {
			pr_debug("sa_list_copy: sa_index_copy failed\n");
			goto err;
		}
		list_add_tail(dst,&(new->entry));
	}
err:
	return error;
}

int sa_list_clear(struct list_head *list)
{
	int error = 0;
	struct list_head *pos = NULL;
	struct sa_index *tmp = NULL;
	struct sa_index *next = NULL;

	if (!list) {
		pr_debug("sadb_list_clear: list is null\n");
		error = -EINVAL;
		goto err;
	}

	list_for_each_safe(pos, next, list){
		tmp = list_entry(pos, struct sa_index, entry);
		if (tmp) {
			list_del(&(tmp->entry));
			sa_index_kfree(&tmp);
		}
	}
err:
	return error;
}

